@@ -600,3 +600,20 @@ on_optionsPrintRuling_activate         (GtkMenuItem     *menuitem,
 void
 on_optionsDiscardCore_activate         (GtkMenuItem     *menuitem,
                                         gpointer         user_data);
+
+void
+on_fontButton_font_set                 (GtkFontButton   *fontbutton,
+                                        gpointer         user_data);
+
+void
+on_optionsLeftHanded_activate          (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsShortenMenus_activate        (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_optionsAutoSavePrefs_activate       (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
